@problem_grid.pro
;========================================================================
;                   PROBLEM GENERATOR MODULE
;
; This module is meant for creating the setup for the 2-D radiative 
; transfer problem of the test cases.
;========================================================================


;------------------------------------------------------------------------
;          HELPER ROUTINE FOR MAIN PROBLEM GENERATOR
; New version: a broken power law possible.
;------------------------------------------------------------------------
function makedens,r,theta,r0,sigma0,hr0,p1,p2,plh,hrmin=hrmin,$
                  hrstore=hrstore,cdens=cdens
thmax = !pi/2.d0
nr=n_elements(r)
nt=n_elements(theta)
rho=fltarr(nr,nt)
hrstore=fltarr(nr)
if not keyword_set(hrmin) then hrmin=0.d0
for ir=0,nr-1 do begin
   if r[ir] le r0 then pl = p1 else pl = p2
   hr = hr0 * (r[ir]/r0)^plh
   hr = sqrt(hr^2+hrmin^2)
   if not keyword_set(cdens) then begin
      rho(ir,*) = 3.98942280401d-1 * sigma0 * (r[ir]/r0)^pl * $
                  exp(-0.5*((thmax-theta(*))/hr)^2 ) / (hr*r[ir])
   endif else begin
      rho(ir,*) = ((!pi/2-theta[*]) lt hr) * sigma0 * $
                  (r[ir]/r0)^pl / (2*hr*r[ir])
   endelse
   hrstore[ir] = hr
endfor
return,rho
end



;------------------------------------------------------------------------
;              PUT MASS ON RANDOM POSITIONS
;------------------------------------------------------------------------
function random_blobs,r,theta,r0,sigma0,hr0,p1,p2,plh,$
             nblob,relsizeblob,istyle,iseed=iseed,hrmin=hrmin,$
             masstot=masstot,sblob=sblob,tq=tq,cdens=cdens,$
             m_random=m_random,r_random=r_random,t_random=t_random
;
; First make a high-resolution 2-D density distribution as if no 
; blobs were used (we use this as the probability function)
;
rin = min(r)
rout= max(r)
nra = 200
nta = 100
ra  = rin * (rout/rin)^(dindgen(nra)/(nra-1.d0))
ta  = 0.5*!pi*dindgen(nta)/(nta-1.d0)
qq  = makedens(ra,ta,r0,sigma0,hr0,p1,p2,plh,hrmin=hrmin,cdens=cdens)
;
; Compute from this the Sigmar2(R)
;
sigma=dblarr(nra)
sigmar2=dblarr(nra)
for ir=0,nra-1 do begin
   sigma[ir]   = 2*integrate(ra[ir]*ta,transpose(qq[ir,*]))
   sigmar2[ir] = sigma[ir]*ra[ir]^2
endfor
;
; Compute the total dust mass of the torus
;
masstorus = diskmass(ra,ta,qq,gastodust=1.)
;
; Make the random positions in r and theta, and the corresponding masses
; The random positions are equally distributed in log(R) and cos(theta). 
; The masses are proportional to rho*r^3 (note: they will be normalized later)
;
case istyle of
   1: begin
      r_random = dblarr(nblob)
      m_random = dblarr(nblob)
      t_random = dblarr(nblob)
      for iblob=0,nblob-1 do begin
         r_random[iblob] = rin * (rout/rin)^(randomu(iseed,/uniform))
         t_random[iblob] = acos(randomu(iseed,/uniform))
         hunt,ra,r_random[iblob],ir,eps_r
         hunt,ta,t_random[iblob],it,eps_t
         dum0 = (1-eps_t)*qq[ir,it]+eps_t*qq[ir,it+1]
         dum1 = (1-eps_t)*qq[ir+1,it]+eps_t*qq[ir+1,it+1]
         dumm = (1-eps_r)*dum0+eps_r*dum1
         m_random[iblob] = (r_random[iblob]^3)*dumm
         print,'mblob = ',m_random[iblob]/1.99d33,' dumm = ',dumm
      endfor
   end
   2: begin
      r_random = dblarr(nblob)
      m_random = dblarr(nblob)
      t_random = dblarr(nblob)
      for iblob=0,nblob-1 do begin
         r_random[iblob] = exp(randomdistr(alog(ra),sigmar2,iseed))
         hunt,ra,r_random[iblob],ir,eps
         if ir lt 0 or ir ge nra then stop
         rh = (1-eps)*qq[ir,*]+eps*qq[ir+1,*]
         t_random[iblob] = acos(randomdistr(cos(ta),rh,iseed))
         m_random[iblob] = 1.d0
      endfor
   end
   10: begin
      r_random = dblarr(nblob)
      m_random = dblarr(nblob)
      t_random = dblarr(nblob)
      for iblob=0,nblob-1 do begin
         r_random[iblob] = exp(randomdistr(alog(ra),sigma,iseed))
         hunt,ra,r_random[iblob],ir,eps
         if ir lt 0 or ir ge nra then stop
         rh = (1-eps)*qq[ir,*]+eps*qq[ir+1,*]
         t_random[iblob] = acos(randomdistr(cos(ta),rh,iseed))
         m_random[iblob] = (r_random[iblob]/ra[0])^2*sin(t_random[iblob])
      endfor
   end
endcase
;
; Now normalize
;
masstot=0.d0
for iblob=0,nblob-1 do begin
   masstot = masstot + m_random[iblob]
endfor
m_random = m_random * (masstorus/masstot)
masstot  = masstorus
;
; Send a message to user
;
tq = m_random * 0.d0
if istyle eq 10 then begin
   for iblob=0,nblob-1 do begin
      rhoavblob = m_random[iblob]/$
           (2*!dpi^2*sblob^2*r_random[iblob]^3*sin(t_random[iblob]))
      tq[iblob] = rhoavblob * sblob * r_random[iblob]
   endfor
endif
;
; Now put the blobs on the grid
;
if n_elements(iseed) eq 0 then iseed=324234
nr=n_elements(r)
nt=n_elements(theta)
rho=dblarr(nr,nt)
rr = rebin(r,nr,nt)
tt = transpose(rebin(theta,nt,nr))
for iblob=0,nblob-1 do begin
   ;;
   ;; Get the random log(R) and cos(theta)
   ;;
   rblob = r_random[iblob]
   tblob = t_random[iblob]
   print,rblob,(!pi/2-tblob)/(!pi/2)
   ;;
   ;; Put the blob on a grid
   ;;
   xx      = (rr-rblob)/rblob
   yy      = (tt-tblob)
   rhoblob = exp(-0.5*(xx^2+yy^2)/relsizeblob^2)
   ;;
   ;; Also add it's mirror image
   ;;
   xx      = (rr-rblob)/rblob
   yy      = (tt-!pi+tblob)
   rhoblob = rhoblob + exp(-0.5*(xx^2+yy^2)/relsizeblob^2)
   ;;
   ;; Normalize the mass of this blob
   ;;
   m = diskmass(r,theta,rhoblob,gastodust=1.)
   rhoblob=rhoblob*(m_random[iblob]/m)
   ;;
   ;; Add this blob to the total rho
   ;; 
   rho = rho + rhoblob
endfor
return,rho
end


;------------------------------------------------------------------------
;                     MAKE OPTICAL DEPTHS
;------------------------------------------------------------------------
function maketau,a,kappa
sz=size(a.rho)
nr=sz(1)
nt=sz(2)
taur=fltarr(nr,nt)
taut=fltarr(nr,nt)
for it=0,nt-1 do begin
    taur(0,it) = 0.d0
    for ir=1,nr-1 do begin
        taur(ir,it) = taur(ir-1,it) + $
            0.5*(a.rho(ir,it)+a.rho(ir-1,it))*kappa * $
            (a.r(ir)-a.r(ir-1))
    endfor
endfor
for ir=0,nr-1 do begin
    taut(ir,0) = 0.d0
    for it=1,nt-1 do begin
        taut(ir,it) = taut(ir,it-1) + $
            0.5*(a.rho(ir,it)+a.rho(ir,it-1))*kappa * $
            (a.theta(it)-a.theta(it-1))*a.r(ir)
    endfor
endfor
return,{taur:taur,taut:taut}
end



;------------------------------------------------------------------------
;                 PROBLEM MAIN FUNCTIONS
;------------------------------------------------------------------------
function problem,nr,nt,rin,rout,r0,hrgrid,hrdisk,plh,sigma0,p1,p2,$
    rrefine=rrefine,hrlg=hrlg,hrmin=hrmin,$
    ntextra=ntextra,hrgmax=hrgmax,nblob=nblob,sblob=sblob,istyle=istyle,$
    iseed=iseed,tq=tq,m_random=m_random,r_random=r_random,t_random=t_random,$
    cdens=cdens
;
; ...Grid refinement near inner edge
;
if not keyword_set(rrefine) then begin
    rrefine={nlevr:3,nspanr:2,nstepr:2}
endif
;
; Make the R-grid
;
r      = make_rgrid(rin,rout,nr,rrefine=rrefine)
;
; Make the Theta-grid
;
theta  = make_tgrid(hrgrid,nt,hrgmax=hrgmax,ntex=ntextra,$
                            hrlg=hrlg,zrefine=zrefine)
;
; The density distribution:
;
if nblob eq 0 or sblob eq 0 then begin
   ;;
   ;; Classical (non-blobby) manner
   ;;
   rho=makedens(r,theta,r0,sigma0,hrdisk,p1,p2,plh,hrmin=hrmin,$
                hrs=hrs,cdens=cdens)
endif else begin
   ;;
   ;; Blobby manner
   ;;
   if nblob eq 0 or sblob eq 0 then begin
      print,'ERROR: If blobby: then define nblob AND sblob'
      stop
   endif
   if n_elements(istyle) eq 0 then istyle=1
   rho=random_blobs(r,theta,r0,sigma0,hrdisk,p1,p2,plh,nblob,sblob,istyle,$
        iseed=iseed,sblob=sblob,tq=tq,m_random=m_random,r_random=r_random,$
        t_random=t_random,cdens=cdens)
endelse
;
; Make the sigma array
;
nnr   = n_elements(r)
nnt   = n_elements(theta)
sigma = dblarr(nnr)
for ir=0,nnr-1 do begin
   if r[ir] le r0 then plsig = p1 else plsig = p2
   sigma[ir] = sigma0 * (r[ir]/r0)^plsig
endfor
;
rr=rebin(r,nnr,nnt)
tt=transpose(rebin(theta,nnt,nnr))
;
return,{r:r,theta:theta,rr:rr,tt:tt,rho:rho,sigma:sigma}
end




;------------------------------------------------------------------------
;                 WRITE PROBLEM FOR RADICAL 
;------------------------------------------------------------------------
pro write_problem_rth,a,abun
nr=n_elements(a.r)
nt=n_elements(a.theta)
nspec=n_elements(abun)
openw,1,'radius.inp'
printf,1,nr
printf,1,' '
for i=1,nr do printf,1,a.r(i-1)
close,1
openw,1,'theta.inp'
printf,1,nt,1
printf,1,' '
for i=1,nt do printf,1,a.theta(i-1)
close,1
openw,1,'dustdens.inp'
printf,1,nspec,nr,nt,1
printf,1,' '
for ispec=1,nspec do begin
   for ir=1,nr do begin
      for it=1,nt do begin
         printf,1,a.rho(ir-1,it-1)*abun(ispec-1)
      endfor
   endfor
   printf,1,' '
endfor
close,1
end





;-----------------------------------------------------------------
;                         PLOT THE GRID
;-----------------------------------------------------------------
pro plotgrid,a,data=data,rmax=rmax,ymax=ymax,ax=ax,az=az,$
           contour=contour,nlevels=nlevels,_extra=_extra
if not keyword_set(rmax) then rmax=max(a.r)
if not keyword_set(data) then data=alog10(a.rho)
if not keyword_set(az) then az=0
if not keyword_set(ax) then ax=90
if not keyword_set(ymax) then ymax=rmax
if not keyword_set(nlevels) then nlevels=100
if not keyword_set(contour) then begin
   surface,data,a.rr*sin(a.tt),a.rr*cos(a.tt),ax=ax,az=az,$
     xrange=[0,rmax],yrange=[0,ymax],_extra=_extra
endif else begin 
   contour,data,a.rr*sin(a.tt),a.rr*cos(a.tt),$
     xrange=[0,rmax],yrange=[0,ymax],nlevels=nlevels,_extra=_extra
endelse
end


;-----------------------------------------------------------------
;                 FIND A FREQUENCY IN A TABLE
;-----------------------------------------------------------------
function find_freq,freq,nu
nf=n_elements(freq)
idx=-100
for inu=0,nf-2 do begin
   if (nu ge freq(inu) and nu le freq(inu+1)) or $
     (nu ge freq(inu+1) and nu le freq(inu)) then begin
      idx = inu
   endif
endfor
if idx lt 0 then begin
   print,"ERROR: nu out of range"
   stop
endif
return,idx
end

;-----------------------------------------------------------------
;                COMPUTE THE MASS OF THE DISK
;-----------------------------------------------------------------
function mass,rhodust,r,t,gastodust=gastodust
if n_elements(gastodust) eq 0 then gastodust=100.
if max(t) le !pi/2 then begin
   imirt=1 
endif else begin 
   imirt=0
   print,'Warning: mirror theta not active'
endelse
nr=n_elements(r)
nt=n_elements(t)
ri=dblarr(nr+1)
ti=dblarr(nt+1)
ri[0]=r[0]
ri[nr]=r[nr-1]
ti[0]=0.
if imirt eq 1 then ti[nt]=!dpi/2 else ti[nt]=!dpi
for ir=1,nr-1 do ri[ir]=sqrt(r[ir]*r[ir-1])
for it=1,nt-1 do ti[it]=0.5*(t[it]+t[it-1])
m=0.d0
for ir=0,nr-1 do begin
   for it=0,nt-1 do begin
      vol = (2.*!dpi/3.)*(ri[ir+1]^3-ri[ir]^3)*abs(cos(ti[it])-cos(ti[it+1]))
      m   = m + vol * rhodust[ir,it]
   endfor
endfor
m=m*gastodust
if imirt eq 1 then m=m*2
return,m
end


;-----------------------------------------------------------------
;                 READ THE DUST OPACITY FILES 
;-----------------------------------------------------------------
function readopac,nr=nr
close,1
if not keyword_set(nr) then nr=1
filename = 'dustopac_'+strcompress(string(nr),/remove_all)+'.inp'
print,"Reading ",filename
openr,1,filename
nf=0
ns=0
readf,1,nf,ns
if nf ge 1 then begin
    cabs = fltarr(nf,ns)
    csca = fltarr(nf,ns)
    dum  = 0.e0
    for k=1,nf do begin
        for is=1,ns do begin
            readf,1,dum
            cabs(k-1,is-1) = dum
        endfor
    endfor
    for k=1,nf do begin
        for is=1,ns do begin
            readf,1,dum
            csca(k-1,is-1) = dum
        endfor
    endfor
    nrtrange=1
    trange=[0.d0,0.d0]
endif else begin
    nf=ns
    ismooth=0
    nrtrange=0
    readf,1,ns,ismooth,nrtrange
    if(ismooth ne 0) then stop   ; No smoothing yet allowed.
    cabs = fltarr(nf,ns,nrtrange)
    csca = fltarr(nf,ns,nrtrange)
    dum  = 0.e0
    trange=fltarr(nrtrange+1)
    for ir=1,nrtrange do begin
        readf,1,a,b
        trange(ir)=b
        for k=1,nf do begin
            for is=1,ns do begin
                readf,1,dum
                cabs(k-1,is-1,ir-1) = dum
            endfor
        endfor
        for k=1,nf do begin
            for is=1,ns do begin
                readf,1,dum
                csca(k-1,is-1,ir-1) = dum
            endfor
        endfor
    endfor
endelse
close,1
file='frequency.inp'
dd=findfile(file,count=count)
if(count le 0) then begin
    print,"Could not find frequency.inp. Taking frequency.dat"
    file='frequency.dat'
    dd=findfile(file,count=count)
    if(count le 0) then begin
       print,"Could not find frequency.dat either"
       stop
    endif
endif
openr,1,file
nnf=0
readf,1,nnf
if nnf ne nf then begin
   print,"ERROR: frequency file has different nr of points as dustopac file"
   stop
endif
freq = fltarr(nf)
wave = fltarr(nf)
for k=1,nf do begin
    readf,1,dum
    freq(k-1) = dum
    wave(k-1) = 2.9979d14 / dum
endfor
close,1
return,{nf:nf,ns:ns,freq:freq,wave:wave,cabs:cabs,csca:csca,nrt:nrtrange,trange:trange}
end
