;==========================================================================
;               MAIN SETUP ROUTINE FOR THE AGN TORUS MODELS
;                 (PARAMETERS ARE IN PROBLEM_PARAMS.PRO)
;==========================================================================
@problem_subroutines.pro
@problem_build.pro
@problem_makeopac.pro
@problem_kurucz.pro
@problem_mixopacities.pro
@problem_files.pro
;
; Natural constants
;
AU     = 1.496d13
cc     = 2.9979d10
RS     = 6.96d10
MS     = 1.99d33
LS     = 3.8525d33     
pc     = 3.08572d18
ss     = 5.6703d-5  
kk     = 1.3807d-16   
mp     = 1.6726d-24  
GG     = 6.672d-8 
TS     = 5.78d3
;
; Other constants
;
mugas  = 2.3
;
; Defaults
;
if n_elements(show) eq 0 then begin
   show   = 1           ; If 1, then show plots of setup
endif
;
; Get date, hostname and current directory
;
spawn,'echo `hostname`',hostname
hostname=strcompress(hostname,/remove_all)
hostname=hostname[0]
cd,current=directory
date=systime()
;
; Open and clear the log file
;
openw,1,'diskmodel.log'
printf,1,'Setup of 2-D Disk Model Run:'
printf,1,'    Date      : '+date
printf,1,'    Hostname  : '+hostname
printf,1,'    Directory : '+directory
printf,1,'-------------------------------------------------'
close,1
;
; Read the parameters
;
@problem_params.pro
;
; Self-consistency checks on parameters
;
if rin ne 0.d0 and tin ne 0.d0 then begin
   print,'PROBLEM: Cannot specify rin AND tin. Must specify'
   print,'         rin OR tin...'
   stop
endif
if mdisk gt 0.d0 and sig0 gt 0.d0 then begin
   print,'PROBLEM: In problem_params.pro one must specify EITHER'
   print,'    the disk mass mdisk, OR the Sigma(rdisk) sig0. Not both!'
   stop
endif
;
; Mix dust opacities
;
if keyword_set(mixspecs) then begin
   sz = size(mixspecs)
   if sz[0] eq 1 then nmix=1 else nmix=sz[2]
   for imix=0,nmix-1 do mixopacities,mixspecs[*,imix],$
                   mixnames[imix],mixabun[*,imix]
endif
;
; Create the opacities and the frequency grid
;
;dustcode = src_dust+'/makedust'
;makeopac,fresmd=fresmd,sil=sil,slb=slb,car=car,ice=ice,frs=frs,$
;         size=size,smin=smin,smax=smax,nsz=nsz,szmix=szmix,pls=pls,$
;         scat=scat,albedo=albedo,allam=allam,alpl=alpl,$
;         dscode=dustcode,dsdata=dustdata
;
useopac,infile,pll,fresmd=fresmd,scat=scat
;
; Read this opacity
;
o      = readopac()
nf     = n_elements(o.freq)
;;;
;;; Create the AGN spectrum (a power law for the moment)
;;;
;;lnu = o.freq^plspec
;;lum = integrate(o.freq,lnu)
;;lnu = lnu*(lagn/lum)
;
; Create AGN spectrum of Granato & Danese (1994)
;
nu        = o.freq
lnu       = nu*0.d0
;
nu1       = 10^(13.5)
nu2       = 10^(14.5)
pl        = 2.0
a         = 1.d0
ii        = where(nu gt nu1 and nu le nu2)
lnu[ii]   = a * (nu[ii]/nu1)^pl
lnu2      = a * (nu2/nu1)^pl
;
nu1       = 10^(14.5)
nu2       = 10^(15.4)
pl        = -0.5
a         = lnu2
ii        = where(nu gt nu1 and nu le nu2)
lnu[ii]   = a * (nu[ii]/nu1)^pl
lnu2      = a * (nu2/nu1)^pl
;
nu1       = 10^(15.4)
nu2       = 10^(16.0)
pl        = -1.0
a         = lnu2
ii        = where(nu gt nu1 and nu le nu2)
lnu[ii]   = a * (nu[ii]/nu1)^pl
lnu2      = a * (nu2/nu1)^pl
;
nu1       = 10^(16.0)
nu2       = 10^(20.0)
pl        = -2.2
a         = lnu2
ii        = where(nu gt nu1 and nu le nu2)
lnu[ii]   = a * (nu[ii]/nu1)^pl
lnu2      = a * (nu2/nu1)^pl
;
lnutot    = integrate(nu,lnu)
lnu       = lnu*(lagn/lnutot)
;
; Write the spectrum
;
openw,1,'starspectrum.inp'
printf,1,nf
printf,1,' '
for inu=0,nf-1 do begin
   printf,1,o.freq[inu],lnu[inu]/1.1965280793d38
endfor
close,1
nu     = o.freq
nulnu  = nu * lnu
;
; Make starinfo.inp
;
openw,1,'starinfo.inp'
printf,1,'1'
printf,1,rstar
printf,1,mstar
printf,1,10.
close,1
;
; Find the opacity at this frequency (wavelength)
;
;lambda = 0.55d0
inupk  = find_freq(o.freq,3d14/lambda)
kappa  = o.cabs(inupk) + o.csca(inupk)
print,"lambda = ",lambda," micron -->  kappa = ",kappa
;
; Do a big loop for the setup. The reason such a loop is required is 
; that if the mass of the disk is given, the Sigma(Rdisk) is not fixed.
; But this value is required in order to compute self-irradiation
; corrections to the radius of the inner rim. If sig0 (==Sigma(Rdisk))
; is given instead (and mdisk=0.d0) then only one loop is done, since
; then the Sigma(R) is fixed.
;
if mdisk gt 0.d0 then sig0=1.d0
a = problem(nr,nt,rin,rout,rdisk,hrgrid,hrdisk,plh,sig0,plsig1,plsig2,$
            rrefine=rrefine,hrmin=hrmin,$
            ntextra=ntex,hrgmax=hrgmax,nblob=nblob,sblob=sblob,$
            iseed=iseed,istyle=istyle,tq=tq,cdens=cdens,$
            m_random=m_random,r_random=r_random,t_random=t_random)
nnr   = n_elements(a.r)
nnt   = n_elements(a.theta)
rr    = rebin(a.r,nnr,nnt)
tt    = transpose(rebin(a.theta,nnt,nnr))
xx    = rr*sin(tt)
yy    = rr*cos(tt)
if mdisk gt 0.d0 then begin 
   md    = mass(a.rho,a.r,a.theta)
   sig0  = mdisk/md
   a.sigma=a.sigma*mdisk/md
   a.rho = a.rho*mdisk/md
   if nblob gt 0 and sblob gt 0.d0 then begin
      m_random = m_random*mdisk/md
      tq       = tq*mdisk/md
      for i=0,n_elements(tq)-1 do print,'r = ',r_random[i]/pc,$
        ' m = ',m_random[i]/MS,$
        ' tau_V = ',tq[i]*kappa
   endif
endif
;
; A-posteriori smoothing of density
;
if drsm ne 0.d0 then begin
   ;;
   ;; Make a smooth curve from 0 to 1, saturating at 1 around
   ;; R=R_0*(1+drsm), i.e. the point where the reduction should stop.
   ;; Using this smooth function we can prevent the reduction from
   ;; being abrupt.
   ;;
   r1 = a.r[0]*(1.d0+drsm)
   x  = (a.r-a.r[0])/(r1-a.r[0])
   q1 = 2.d1
   q2 = 1.d0
   z  = -alog10(exp(-x*q1)+10^(-q2))/q2   ; Make smooth curve
   z  = 1-(1-z)/(1-z[0])                  ; Make sure z[0]==0
   ;;
   ;; Now reduce the density.
   ;;
   if not keyword_set(rdxsm) then rdxsm=1d-2
   for ir=0,nnr-1 do begin
      x = 1.d0-z[ir]
      redux=rdxsm^x
      a.rho(ir,*)=a.rho(ir,*)*redux
      a.sigma(ir)=a.sigma(ir)*redux
   endfor
endif
;
; Compute the mass
;
md = mass(a.rho,a.r,a.theta)
print,'Mdisk = ',md/MS
;
; Make optical depth array
;
t=maketau(a,kappa)
;
; Write a message
;
ddr=a.r(nnr-1)/a.r(nnr-2)-1.d0
print,'TAUR  = ',t.taur(nnr-1,nnt-1)
print,'DR/R  = ',ddr
;
; Do tests
;
if ddr lt 0.05 then begin
   print,'Radial grid too finely spaced... This will take too much'
   print,'computational time... Are you sure this is okay? (Type 1)'
   read,i
   if i ne 1 then stop
endif
if ddr gt 0.15 then begin
   print,'Radial grid too coarse! Are you sure this is okay? (Type 1)'
   read,i
   if i ne 1 then stop
endif
;
; Check the optical depth at the inner edge
;
tr0 = t.taur(1,nnt-1)
if tr0 gt 1.d0 then begin
   print,'Optical depth error! Inner grid cell at equator optically thick'
   print,' Is this okay? (type 1)'
   read,i
   if i ne 1 then stop
endif
;
; Write the stuff to the input files
;
write_problem_rth,a,[1.0]
;
; Write the radmc.inp file
;
write_radmc,nphot=nphot,/ifast
;
; Write the raytrace.inp for the ray-tracing (spectra and images)
;
write_raytrace,nextra=nextra,nrref=nrref
;
; Write a data file with the names simser,simnr,simweb
;
openw,1,'diskmodel.names'
printf,1,simser
printf,1,simnr
printf,1,simweb
close,1
;
; Make plots and write info
;
if show eq 1 then begin
   tt = t
   pos = [0.1,0.55,0.95,0.95]
   xtitle = 'R [AU]'
   ytitle = '!4p!X/2-!4H!X = H/R'
   nnt = n_elements(a.theta)
   window,1,xsize=800,ysize=700
   !p.multi(2)=2
;   surface,tt.taur gt 1,r,theta,/xl,ax=50,/xs,/ys
   surface,tt.taur gt 1,a.r/au,!pi/2-a.theta,/xl,ax=90,az=0,/xs,/ys,position=pos,xtitle=xtitle,ytitle=ytitle
   contour,tt.taur gt 1,a.r/au,!pi/2-a.theta,/xl,/xs,/ys,position=pos,/noerase,/fill,$
          title='Grid and the !4s!X!DV!N>1 region'
;   surface,tt.taur gt 1,r/au,!pi/2-theta,xtitle=xtitle,ytitle=ytitle,/xl,ax=90,az=0,/xs,/ys,position=pos,/noerase
   plot,tt.taur(0:20,nnt-1),yrange=[0,6],psym=6,yminor=1,yticks=6,$
            xtitle='Radial grid points from inner edge',$ 
            ytitle='!4s!X!DV!N at midplane'
   oplot,tt.taur(0:20,nnt-1)
   !p.multi(2)=1
endif
;
; Do the run
;
spawn,'nice '+bindir+'radmc'
spawn,'nice '+bindir+'raytrace spectrum incl 30'
;
end
