;====================================================================
;
;      IDL ROUTINES FOR READING/ANALYZING RADMC/RAYTRACE OUTPUT
;
;                    NEW VERSION MARCH 2007
;
;                    C.P. Dullemond (2007)
;
; read_dustdens()      Reads the dust density structure, and also
;                      computes the Sigma_dust_i(R) for each species,
;                      and computes the total dust mass and disk mass.
;
; read_dusttemp()      Reads the dust temperature structure that is
;                      located in the file dusttemp_final.dat. This
;                      is the result of the RADMC calculation.
;
; read_spectrum()      Reads the spectrum that comes out of a run
;                      of RAYTRACE and is stored in spectrum.dat
;
; plot_spectrum        Subroutine that plots the SED read by the 
;                      read_spectrum() routine.
;
; read_obsspec()       Reads in observed spectra in some format. This
;                      is presumably not used by those who have their
;                      observed SEDs in their own formats.
;
; read_photstat()      Read the photon statistics file (photstat.info)
;                      to see how many MC photon packages pass by in
;                      every cell.
;
;====================================================================

;--------------------------------------------------------------------
;              A FUNCTION USED BY THE ROUTINES BELOW
;--------------------------------------------------------------------
function sigmamass,rhodust,r,t,imirt
nr=n_elements(r)
nt=n_elements(t)
ri=dblarr(nr+1)
ti=dblarr(nt+1)
ri[0]=r[0]
ri[nr]=r[nr-1]
ti[0]=0.
if imirt eq 1 then ti[nt]=!dpi/2 else ti[nt]=!dpi
for ir=1,nr-1 do ri[ir]=sqrt(r[ir]*r[ir-1])
for it=1,nt-1 do ti[it]=0.5*(t[it]+t[it-1])
m=0.d0
sigma=dblarr(nr)
for ir=0,nr-1 do begin
   surf = !dpi * ( ri[ir+1]^2 - ri[ir]^2 )
   for it=0,nt-1 do begin
      vol       = (2.*!dpi/3.)*(ri[ir+1]^3-ri[ir]^3)*abs(cos(ti[it])-cos(ti[it+1]))
      m         = m + vol * rhodust[ir,it]
      sigma[ir] = sigma[ir] + vol * rhodust[ir,it] / surf
   endfor
endfor
if imirt eq 1 then begin
   m=m*2
   sigma=sigma*2
endif
return,{sigma:sigma,m:m}
end


;====================================================================
;                         READ FUNCTIONS
;====================================================================


;--------------------------------------------------------------------
;                    READ THE GRID INFORMATION
;--------------------------------------------------------------------
function read_grid,dat=dat
nr=0
ntheta=0
nthsmall=0
nfr=0
imirt=0
close,1
;
; Read the radius data
;
file='radius.inp'
dd=findfile(file,count=count)
if keyword_set(dat) then count=0
if(count le 0) then begin
    file='radius.dat'
    dd=findfile(file,count=count)
    if(count le 0) then stop
endif
openr,1,file
readf,1,nr
r=dblarr(nr)
readf,1,r
close,1
;
; Read the theta data
;
file='theta.inp'
dd=findfile(file,count=count)
if keyword_set(dat) then count=0
if(count le 0) then begin
    file='theta.dat'
    dd=findfile(file,count=count)
    if(count le 0) then stop
endif
openr,1,file
readf,1,ntheta,imirt
nthsmall=ntheta
theta=dblarr(ntheta)
readf,1,theta
close,1
if imirt eq 1 then begin
   thbk=dblarr(2*ntheta)
   thbk(0:ntheta-1) = theta(*)
   thbk(ntheta:2*ntheta-1) = !pi - rotate(theta(*),2)
   theta = thbk
   ntheta = ntheta*2
endif
;
; Now read the frequencies
;
file='frequency.inp'
dd=findfile(file,count=count)
if keyword_set(dat) then count=0
if(count le 0) then begin
    file='frequency.dat'
    dd=findfile(file,count=count)
    if(count le 0) then begin
       print,'Could not find frequency.inp/.dat!!'
       nu=[1.0,1.0]
       nnu=0
    endif
endif
filename=file
str=findfile(filename)
if(str(0) eq filename) then begin
    openr,1,filename
    nnu=1
    readf,1,nnu
    nu=dblarr(nnu)
    dum=dblarr(nnu)
    readf,1,dum
    nu(*) = dum(*)
    close,1
endif else begin
    nu=[0.d0,0.d0]
    nnu=0
endelse
;
return,{nr:nr,ntheta:ntheta,nthsmall:nthsmall,nnu:nnu,r:r,theta:theta,nu:nu}
end


;-----------------------------------------------------------------
;                   READ DUST DENSITIES
;-----------------------------------------------------------------
function read_dustdens,ext=ext,dat=dat
mass = 1

grid = read_grid()
dum=0

nrspec=0
sizer=0
sizet=0
imirt=0

filebase='dustdens'
if n_elements(ext) gt 0 then begin
   filebase=filebase+'_'+strcompress(string(ext),/remove_all)
endif
if not keyword_set(dat) then begin
   filename=filebase+'.inp' 
endif else begin
   filename=filebase+'.dat' 
endelse
dd=findfile(filename,count=count)
if(count eq 0) then begin
   filename=filebase+'.dat'
   dd=findfile(filename,count=count)
   if(count eq 0) then stop
endif
openr,1,filename
readf,1,nrspec,sizer,sizet,imirt
sizetbig=sizet*(1+imirt)
dens=dblarr(sizer,sizetbig,nrspec)
idum=0
for ispec=1,nrspec do begin
    for ir=1,sizer do begin
        for it=1,sizet do begin
            dum = 0.d0
            readf,1,dum
            dens(ir-1,it-1,ispec-1) = dum
            if(imirt eq 1) then begin
                itt = 2*sizet + 1 - it
                dens(ir-1,itt-1,ispec-1) = dum
            endif
        endfor
    endfor
endfor
close,1
rr=rebin(grid.r,grid.nr,grid.ntheta)
if grid.ntheta gt 1 then $
   tt=transpose(rebin(grid.theta,grid.ntheta,grid.nr)) $ 
   else tt=0
close,1
if keyword_set(mass) then begin
   sigmadust=dblarr(grid.nr,nrspec)
   massdust=dblarr(nrspec)
   for ispec=0,nrspec-1 do begin
      qq=sigmamass(dens[*,*,ispec],grid.r,grid.theta,0)
      sigmadust[*,ispec]=qq.sigma
      massdust[ispec]=qq.m
   endfor
endif else begin
   sigmadust=-1
   massdust=-1
endelse
return,{nr:grid.nr,ntheta:grid.ntheta,nnu:grid.nnu,$
        r:grid.r,theta:grid.theta,nu:grid.nu(*),$
        nspec:nrspec,rr:rr,tt:tt,rho:dens,$
        sigmadust:sigmadust,massdust:massdust}
end


;-----------------------------------------------------------------
;                   READ DUST TEMPERATURES 
;-----------------------------------------------------------------
function read_dusttemp,ext=ext,all=all
grid = read_grid()
niter=0
dum=0
nrsizemax=0
openr,1,'dusttemp.info'
readf,1,niter
readf,1,dum
readf,1,nrsizemax
readf,1,lastsave
close,1

iter=0
nrspec=0
sizer=0
sizet=0
imirt=0

if not keyword_set(all) then begin
   if(niter eq -2) then niter='final'
   if n_elements(ext) gt 0 then niter=ext
   filename='dusttemp_'+strcompress(string(niter),/remove_all)+'.dat'
endif else begin
   filename=['']
   for i=1,lastsave do begin
      filename=[filename,'dusttemp_'+strcompress(string(i),/remove_all)+'.dat']
   endfor
   if niter eq -2 then begin
      filename=[filename,'dusttemp_final.dat']
   endif
   filename=filename(1:*)
endelse

nn=n_elements(filename)

for in=1,nn do begin
   openr,1,filename[in-1]
   readf,1,nrspec,sizer,sizet,imirt
   sizetbig=sizet*(1+imirt)
   if in eq 1 then begin
      temp=dblarr(sizer,sizetbig,nrsizemax,nrspec,nn)
      nrgrainsize=intarr(nrspec)
   endif
   idum=0
   for ispec=1,nrspec do begin
      readf,1,idum
      nrgrainsize(ispec-1)=idum
      for isize=1,nrgrainsize(ispec-1) do begin
         for ir=1,sizer do begin
            for it=1,sizet do begin
               dum = 0.d0
               readf,1,dum
               temp(ir-1,it-1,isize-1,ispec-1,in-1) = dum
               if(imirt eq 1) then begin
                  itt = 2*sizet + 1 - it
                  temp(ir-1,itt-1,isize-1,ispec-1,in-1) = dum
               endif
            endfor
         endfor
      endfor
   endfor
   close,1
endfor
rr=rebin(grid.r,grid.nr,grid.ntheta)
if grid.ntheta gt 1 then $
   tt=transpose(rebin(grid.theta,grid.ntheta,grid.nr)) $ 
   else tt=0
close,1
;
; If the starinfo file is there, then read this
;
dd=findfile('starinfo.inp',count=count)
if(count gt 0) then begin
   openr,1,'starinfo.inp'
   readf,1,iformat
   readf,1,rstar
   readf,1,mstar
   readf,1,tstar
   close,1
   tthin = sqrt(0.5d0*rstar/rr)*tstar   
endif else begin
   tthin=0.d0
endelse
;
return,{nr:grid.nr,ntheta:grid.ntheta,nnu:grid.nnu,$
        r:grid.r,theta:grid.theta,nu:grid.nu(*),$
        rr:rr,tt:tt,nspec:nrspec,nsize:nrgrainsize,$
        temp:temp,tthin:tthin}
end



;--------------------------------------------------------------------
;                      READ PHOTON STATISTICS
;--------------------------------------------------------------------
function read_photstat

grid = read_grid()

nr=0
nt=0
iformat=0

openr,1,'photstat.info'
readf,1,iformat
if iformat ne 1 then stop
readf,1,nr,nt
ntbig=nt*2
stat=dblarr(nr,ntbig)
idum=0
for it=1,nt do begin
   for ir=1,nr do begin
      readf,1,dum
      stat(ir-1,it-1) = dum
      itt = 2*nt + 1 - it
      stat(ir-1,itt-1) = dum
    endfor
endfor
close,1
rr=rebin(grid.r,grid.nr,grid.ntheta)
if grid.ntheta gt 1 then $
   tt=transpose(rebin(grid.theta,grid.ntheta,grid.nr)) $ 
   else tt=0
return,{nr:grid.nr,ntheta:grid.ntheta,$
        r:grid.r,theta:grid.theta,$
        rr:rr,tt:tt,stat:stat}
end


;--------------------------------------------------------------------
;                      READ THE DIFFUSION INFO
;--------------------------------------------------------------------
function read_diffinfo

grid = read_grid()

nr=0
nt=0
iformat=0

openr,1,'diffusion.info'
readf,1,iformat
if iformat ne 1 then stop
readf,1,nr,nt
ntbig=nt*2
info=dblarr(nr,ntbig)
idum=0
for it=1,nt do begin
   for ir=1,nr do begin
      readf,1,dum
      info(ir-1,it-1) = dum
      itt = 2*nt + 1 - it
      info(ir-1,itt-1) = dum
    endfor
endfor
close,1
rr=rebin(grid.r,grid.nr,grid.ntheta)
if grid.ntheta gt 1 then $
   tt=transpose(rebin(grid.theta,grid.ntheta,grid.nr)) $ 
   else tt=0
return,{nr:grid.nr,ntheta:grid.ntheta,$
        r:grid.r,theta:grid.theta,$
        rr:rr,tt:tt,info:info}
end



;=================================================================
;                     SPECTRUM PLOT ROUTINES     
;=================================================================

;-----------------------------------------------------------------
;                        READ THE SPECTRUM
;-----------------------------------------------------------------
function read_spectrum,ext=ext,file=file,dpc=dpc
fr_units=0
;
; Message of warning:
;
print,'NOTE: Users of earlier versions of analyze.pro:'
print,'      The spectrum array now goes from 0 to nf-1,'
print,'      instead of 1:nf as in earlier versions!'
;
; Read possible dust information
;
;z=findfile('dustopac.inp',count=count)
;dust=count
;if dust gt 0 then begin
;    print,"Found dustopac.inp, so assuming dust spectrum"
;    fr_units = -1
;endif
;
; From now on, always use fr_units=-1 (dust RT results)
;
fr_units = -1
;
; Read the total spectrum
;
if not keyword_set(file) then begin
   if n_elements(ext) eq 0 then begin
      filename='spectrum.dat'
   endif else begin
      filename='spectrum_'+strcompress(string(ext),/remove_all)+'.dat'
   endelse
endif else begin
   filename=file
endelse
openr,1,filename
nrfr=1
readf,1,nrfr
dum=dblarr(2,nrfr)
freq=dblarr(nrfr)
spectrum=dblarr(nrfr)
readf,1,dum
freq(0:nrfr-1) = dum(0,0:nrfr-1)
spectrum(0:nrfr-1) = dum(1,0:nrfr-1)
close,1
;
; If dpc is specified, then the spectrum is not the standard d=1pc
; spectrum, but in fact is a spectrum at a given distance (for instance,
; if you want to use the new option of RADICAL to include an aperture,
; then RADICAL returns the actual spectrum at d=dpc distance). In order
; not to confuse the plot_spectrum() routine one can rescale it back
; to a spectrum at 1pc by specifying dpc here in this read routine. 
;
if keyword_set(dpc) then spectrum=spectrum*dpc^2
;
; Return all...
;
a={spectrum:spectrum,nfr:nrfr,freq:freq,fr_units:fr_units,obs:0}
return,a
end


;-----------------------------------------------------------------
;                   READ AN OBSERVED SPECTRUM
;-----------------------------------------------------------------
function read_obsspec,file,hz=hz,micron=micron,jy=jy,erg=erg,$
      nufnu=nufnu,npt=npt
cc   = 2.9979d10
fr_units = 0
if (not keyword_set(hz)) and (not keyword_set(micron)) then begin
   print,"Must specify either /hz or /micron"
   return,0
endif
if (not keyword_set(jy)) and (not keyword_set(erg)) $
    and (not keyword_set(nufnu)) then begin
   print,"Must specify either /jy or /erg or /nufnu"
   return,0
endif
openr,1,file
if not keyword_set(npt) then readf,1,npt
data=dblarr(2,npt)
readf,1,data
close,1
freq=dblarr(npt)
spectrum=dblarr(npt)
if keyword_set(hz) then begin
   freq(*) = data(0,0:npt-1)
endif 
if keyword_set(micron) then begin
   freq(*) = cc / (1d-4*data(0,0:npt-1))
   fr_units = -1
endif
if keyword_set(erg) then begin
   spectrum(*) = data(1,0:npt-1)
endif
if keyword_set(jy) then begin
   spectrum(*) = 1d-23 * data(1,0:npt-1)
endif
if keyword_set(nufnu) then begin
   spectrum(*) = data(1,0:npt-1) / freq(npt)
endif
a={spectrum:spectrum,nfr:npt,freq:freq,fr_units:fr_units,obs:1}
return,a
end



;-----------------------------------------------------------------
;                       PLOT SPECTRUM 
;
; This routine plots the spectrum read in by read_spectrum().
;
; OPTIONS:
;   lnu             = 1 --> L_nu (default nu*L_nu)
;   fnu             = 1 --> F_nu (instead of nu*F_nu)
;   dpc             = Distance of observer in units of parsec
;                   = 0 (or undef) --> Lum = Total Lum in erg/s
;                     instead of erg / s cm^2
;   lsun            = Plot y-axis in units of Lsun
;   ylin            = Plot y-axis linearly
;   jy              = Plot y-axis in Jansky
;   
;-----------------------------------------------------------------
pro plot_spectrum,a,ev=ev,kev=kev,hz=hz,micron=micron,$
        lnu=lnu,fnu=fnu,nulnu=nulnu,nufnu=nufnu,$
        dpc=dpc,xlg=xlg,oplot=oplot,jy=jy,lsun=lsun,$
        itheta=itheta,ldat=ldat,ylin=ylin,lum=lum,$
        thick=thick,xthick=xthick,ythick=ythick,$
        charthick=charthick,charsize=charsize,_extra=_extra
if not keyword_set(itheta) then itheta=0
fr_units = a.fr_units
if keyword_set(micron) then begin
    fr_units = -1
endif
if keyword_set(hz) then begin
    fr_units = 0
endif
if keyword_set(ev) then begin
    fr_units = 1
endif
if keyword_set(kev) then begin
    fr_units = 2
endif
case fr_units of
    -1: begin
        xcoord = 2.9979d14 / a.freq(*)
        xtitle = '!4k!X [!4l!Xm]'
    end
    0: begin
        xcoord = a.freq(*)
        xtitle = '!4m!X [Hz]'
    end
    1: begin
        xcoord = 4.13568842841d-15 * a.freq
        xtitle = '!4m!X [eV]'
    end
    2: begin
        xcoord = 4.13568842841d-18 * a.freq
        xtitle = '!4m!X [KeV]'
    end
endcase

if keyword_set(dpc) then rpc=dpc
;
; Plot nuFnu or Fnu (same with Lnu)? And what about Fnu vs Lnu?
;
sed=1
ylum=0
if keyword_set(jy) then begin
   sed=0
endif
if keyword_set(fnu) then begin
   ylum=0
   sed=0
endif
if keyword_set(lnu) then begin
   ylum=1
   sed=0
endif
if keyword_set(nufnu) then begin
   ylum=0
   sed=1
endif
if keyword_set(nulnu) then begin
   ylum=1
   sed=1
endif
if keyword_set(lum) then ylum=1 
if keyword_set(jy) then begin
   ylum=0
   ;sed=0
endif
if keyword_set(lsun) then begin
   ylum=1
   sed=1
endif
;
; If the distance is not given, then for an observed spectrum
; we can only plot the flux, while for a computed spectrum we
; can only plot the luminosity
;
if a.obs eq 0 then begin
   if not keyword_set(rpc) and keyword_set(jy) then begin
      print,"Cannot use Jy units of theoretical spectrum if dpc not known."
      return
   endif
   if not keyword_set(rpc) and ylum eq 0 then begin
      ylum=1
      print,"Do not know distance to source. Plotting Luminosity instead."
   endif
endif else begin
   if not keyword_set(rpc) and ylum eq 1 then begin
      ylum=0
      print,"Do not know distance to source. Plotting Flux instead."
   endif
endelse
;
; Which plot to make? Lum or flux?
;
if ylum eq 0 then begin
   ;
   ; Plot spectrum as flux at a certain distance
   ;
   if a.obs eq 0 then begin
      rpc=1.d0*rpc
      distfact = 1.d0/ (rpc^2)
   endif else begin
      distfact = 1.d0
   endelse
   if not keyword_set(jy) then begin
      if sed eq 0 then begin
         lumfact=1.d0
         ytitle='F!I!4m!X!N  [erg cm!E-2!N Hz!E-1!N s!E-1!N]'
      endif else begin
         lumfact=a.freq(*)
         ytitle='!4m!XF!I!4m!X!N [erg cm!E-2!N s!E-1!N]'
      endelse
   endif else begin
      if sed eq 0 then begin
         lumfact=1d+23
         ytitle='F!I!4m!X!N [Jy]'
      endif else begin
         lumfact=1d+23*a.freq(*)
         ytitle='!4m!XF!I!4m!X!N [JyHz]'
      endelse
   endelse
endif else begin
   ;
   ; Plot spectrum as luminosity
   ;
   if a.obs eq 0 then begin
      distfact = 1.1965280793d38 ; = (1 parsec)^2 = 1.19d38 cm^2
   endif else begin
      rpc=1.d0*rpc
      distfact = rpc^2 * 1.1965280793d38
   endelse
   if sed eq 0 then begin
      lumfact=1.d0
      ytitle='L!I!4m!X!N  [erg Hz!E-1!N s!E-1!N]'
   endif else begin
      if not keyword_set(lsun) then begin
         lumfact=a.freq(*)
         ytitle='!4m!XL!I!4m!X!N [erg s!E-1!N]'
      endif else begin
         lumfact=a.freq(*)*2.5956986d-34
;         ytitle='!4m!XL!I!4m!X!N [Lsun]'
         ytitle='!4m!XL!I!4m!X!N [L!D!9n!X!N]'
      endelse
   endelse
endelse
if not keyword_set(xlg) then begin
    xlog=1
endif else begin
    xcoord=alog10(xcoord)
    xtitle = 'log '+xtitle
    xlog=0
endelse
if not keyword_set(ylin) then begin
    ylog=1
endif else begin
    ylog=0
endelse
if not keyword_set(oplot) then begin
    plot,xcoord,distfact*lumfact*a.spectrum(*,itheta),$
         xtitle=xtitle,ytitle=ytitle,$
         xlog=xlog,ylog=ylog,thick=thick,xthick=xthick,ythick=ythick,$
        charthick=charthick,charsize=charsize,_extra=_extra
endif else begin
    oplot,xcoord,distfact*lumfact*a.spectrum(*,itheta),$
         thick=thick,_extra=_extra
endelse
end




